-- Copyright (C) 2009  Carlo A. Furia
--                     caf@inf.ethz.ch  http://se.inf.ethz.ch/people/furia/
--                     
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.


class
	INVARIANT_GENERATION


inherit
	ARGUMENTS


create
	make




feature -- Creation


	make is
			-- process command-line arguments
		local
			new_invariants: INTEGER
			stopwatch: STOP_WATCH
		do
			print ("This is GIN-PINK --- Generation of Invariants by PostconditIoN weaKening")
			create stopwatch
			create token_codes
			if argument_count = 0 then
				print ("%N" + "No input file. Try " + Command_name
						 + " --help for summary of usage.%N")
			else
				process_command_line_arguments
				if run_boogie and not {PLATFORM}.is_windows then
					print ("%N" + "Option --run-boogie is unsupported under your platform."
							 + " Ignoring it.")
					run_boogie := False
				end
				if valid_input_file then
					stopwatch.start
					parse_boogie_file
					if boogie_file_parsed then
						generate_invariants_on_file
						if invariant_file_created then
						print ("%N" + "Invariant generation succeeded.")
						generate_mutated_programs_on_file
							if mutated_files_created then
								print ("%N" + "Mutated files generation succeeded.")
								if run_boogie then
									-- run boogie
									new_invariants := lets_boogie
									if new_invariants > 0 then
										print ("%N%N" + "Found "); print (new_invariants)
										print (" new invariants! :-)")
									else
										print ("%N%N" + "Found no new invariants. :-(")
									end
								end
							else
								print ("%N" + "Mutated files generation failed.")
							end
						else
							print ("%N" + "Invariant generation failed")
						end
					else
						print ("%N" + "Input file: " + boogie_filename +
								 " does not parse correctly")
					end
					stopwatch.stop
					print ("%N" + "Total running time (sec.): "); print(stopwatch.seconds)
				end
				print ("%N")
			end
			if {PLATFORM}.is_windows then
				print ("%N" + "Press Enter to finish the execution...")
				io.read_character
			end
		end



feature -- Options, status, and error reporting

	-- boogie_filename can be opened
	valid_input_file: BOOLEAN

	-- a bunch of flags for command-line options
	all_variable_subsets: STRING
	reference_based_substitution: BOOLEAN
	try_previous_values: BOOLEAN
	run_boogie: BOOLEAN

			-- parameters to launch Boogie with
	boogie_parameter: STRING
			-- filename to be parsed
	boogie_filename: STRING
			-- extension of Boogie filenames
	boogie_extension: STRING is ".bpl"
			-- boogie_filename has been parsed correctly
	boogie_file_parsed: BOOLEAN

			-- name of procedure to be analyzed
	      -- (empty string if the first is chosen)
	procedure_name: STRING

			-- suffix for file containing invariants
	invariant_file_suffix: STRING is ".invariants"
			-- invariant files have been successfully created and dumped on disk
	invariant_file_created: BOOLEAN

			-- suffix for mutated file(s) containing invariants
	mutated_file_suffix: STRING is "_CANDIDATE_INVARIANT_"
			-- mutated Boogie files with the new candidate invariants
			-- have been successfully created and dumped on disk
	mutated_files_created: BOOLEAN

			-- command string to launch Boogie
	boogie_command_name: STRING is "boogie"


			-- prefix for names of non-verified files
	failed_name: STRING is "_FAILED_"







feature -- Invariant files creation and execution

	mutated_files_names: LINKED_LIST [STRING]
		-- names of all files with mutated programs
		-- (is Void before first successfull call to generated_mutated_programs_on_file)

	parse_boogie_file is
			-- parse boogie_filename
			-- and set boogie_file_parsed to True upon sucess
		do
			create boogie_parsed_file.make (boogie_filename)
			if boogie_parsed_file.parse_tree = Void then
				-- something went wrong while parsing
				boogie_file_parsed := False
			else
				boogie_file_parsed := True
			end
		end



	generate_invariants_on_file is
			-- generate all invariants according to current options
			-- from currently parsed file, and dump them on file
			-- invariant_file_created is set to True upon success
		require
			boogie_file_parsed
		local
			out_file_string: STRING
			current_navigator: PRODUCTION_CLASS_NAVIGATOR

		do
			invariant_file_created := False
			create weakener.make (boogie_parsed_file.parse_tree)
			weakener.choose_procedure_name (procedure_name)

			-- set options on weakener
			if all_variable_subsets /= Void and then
			   all_variable_subsets.is_integer and then
			   all_variable_subsets.to_integer > 0 then
				weakener.do_multiple_variables_at_a_time (all_variable_subsets.to_integer)
			else
				weakener.do_one_variable_at_a_time
			end
			if reference_based_substitution then
				weakener.base_substitution_on_reference
			else
				weakener.base_substitution_on_name
			end
			if try_previous_values then
				weakener.try_previous_values
			else
				weakener.no_previous_values
			end

			-- generate all invariants
			weakener.generate_candidate_invariants
			if weakener.all_candidate_invariants /= Void then
				-- invariants correctly generated
				create out_file_string.make (30)
				from weakener.all_candidate_invariants.start
				until weakener.all_candidate_invariants.after
				loop
					create current_navigator.make (weakener.all_candidate_invariants.item)
					current_navigator.start (token_codes.Expr)
					out_file_string.append (current_navigator.print_leaves + "%N")
					weakener.all_candidate_invariants.forth
				end
				-- dump on file
				create current_file.make (boogie_filename + invariant_file_suffix)
				if current_file.is_creatable then
					current_file.open_write
					current_file.put_string (out_file_string)
					current_file.close
					invariant_file_created := True
				end
			end
		end




	generate_mutated_programs_on_file is
			-- generate all mutated programs with generated invariants
			-- mutated_files_created is set to True upon success
		require
			invariant_file_created
		local
			current_navigator: POSTCONDITION_WEAKENER
			out_file_string: STRING
			suffix_starts_at: INTEGER
			current_filename: STRING
		do
			create current_navigator.make (boogie_parsed_file.parse_tree)
			current_navigator.choose_procedure_name (procedure_name)
			current_navigator.set_current_procedure
			create mutated_files_names.make

			from weakener.all_candidate_invariants.start
			until weakener.all_candidate_invariants.after
			loop
				current_navigator.add_invariant_to_all_while_loops (weakener.all_candidate_invariants.item)
				create out_file_string.make (100)
				current_navigator.start (token_codes.Program)
				out_file_string.append (current_navigator.print_leaves + "%N")
				-- dump_on_file
				create current_filename.make (10)
				suffix_starts_at := boogie_filename.substring_index (boogie_extension, 1)
				if suffix_starts_at > 1 then
					current_filename.append (boogie_filename.substring (1, suffix_starts_at - 1))
				else
					current_filename.append (boogie_filename)
				end
				current_filename.append (mutated_file_suffix)
				current_filename.append
				( padded_integer_as_string
				  (weakener.all_candidate_invariants.index,
					weakener.all_candidate_invariants.count))
				current_filename.append (boogie_extension)
				create current_file.make (current_filename)
				if current_file.is_creatable then
					current_file.open_write
					current_file.put_string (out_file_string)
					current_file.close
					mutated_files_names.extend (current_filename)
				end

				-- output some advancement indicator
				print ("%N" + "Generated mutation # ")
				print (weakener.all_candidate_invariants.index)
				print (" --- ")
				print (weakener.all_candidate_invariants.count
						 - weakener.all_candidate_invariants.index)
				print (" more to go.")

				-- go to next invariant
				weakener.all_candidate_invariants.forth
				-- remove generated invariants
				current_navigator.remove_invariant_from_all_while_loops
			end
			mutated_files_created := True
		end




feature {NONE} -- Implementation

	current_file: PLAIN_TEXT_FILE

	boogie_parsed_file: BOOGIEPL_PARSER

	weakener: POSTCONDITION_WEAKENER

	token_codes: BOOGIEPL_TOKEN_CODES



	input_file_exists: BOOLEAN is
		-- check if a file with boogie_filename exists and can be
			-- opened in read mode
		do
			create current_file.make (boogie_filename)
			Result := current_file.exists and current_file.is_access_readable
		end



	options_go_wrong is
			-- print a message for mismatched command-line options
		do
			print ("%N" + Command_name + ": incorrect command line options. Try "
					 + Command_name + " --help")
		end


	print_help_notice is
			-- print a notice about the usage of the command
		do
			print ("%N" + "Usage: " + Command_name + " [options] [Boogie input filename]")
			print ("%N" + "where [options] is one or more of the following.")
			print ("%N"+ "--procedure-name=NAME : work on procedure NAME (the default is the first procedure in the file)")
			print ("%N"+ "--all-variable-subsets=N :  try every possible combination of replacements for up to N variables")
			print ("%N"+ "--uncouple :  treat separately every occurrences of the same variable")
			print ("%N"+ "--age :  try substituting values of variables one iteration in the past")
			print ("%N"+ "--run-boogie :  run Boogie on generated all files and remove those all of whose invariants fail")
			print ("%N"+ "--boogie-parameter=%"PARAMETERS%" :  pass string PARAMETERS as parameters to Boogie")
			print ("%N" + "--help :  display this notice")
		end



	process_command_line_arguments is
			-- process command line options
		local
			parameter_number: INTEGER
		do
			set_option_sign ('-')
			create boogie_parameter.make (5)

			-- get last argument as input filename
			create boogie_filename.make_from_string (argument_array [argument_count])

			if index_of_word_option ("-help") /= 0 then
				print_help_notice
				valid_input_file := False
			else
				if not input_file_exists then
					print ("%N" + "Can't open input file: " + boogie_filename)
					valid_input_file := False
				else
					valid_input_file := True
					-- gather various command line options
					all_variable_subsets := coalesced_word_option_value ("-all-variable-subsets=")
					reference_based_substitution:= index_of_word_option ("-uncouple") /= 0
					try_previous_values:= index_of_word_option ("-age") /= 0
					run_boogie:= index_of_word_option ("-run-boogie") /= 0

					procedure_name := coalesced_word_option_value ("-procedure-name=")

					parameter_number := index_of_beginning_with_word_option ("-boogie-parameter=%"")
					if parameter_number > 0 and argument_array[parameter_number].ends_with("%"") then
						boogie_parameter := argument_array[parameter_number]
						-- remove "--boogie-parameter=" and final character "="
						boogie_parameter := boogie_parameter.substring
						(1 + boogie_parameter.index_of ('=', 1), boogie_parameter.count - 1)
					end
				end
			end
		end



	padded_integer_as_string (an_integer: INTEGER; largest: INTEGER): STRING is
			-- return a string representation of `an_integer' with leading 0's
			-- such that the total number of digits is the same as those
			-- of `largest'
		require
			an_integer <= largest
			an_integer >= 0
			largest >= 0
		local
			largest_to_string: STRING
			padding: STRING
		do
			create largest_to_string.make_empty
			largest_to_string.append_integer (largest)
			create Result.make (largest_to_string.count)
			Result.append_integer (an_integer)
			create padding.make_filled ('0', largest_to_string.count - Result.count)
			Result.prepend (padding)
		end




	lets_boogie: INTEGER is
			-- run Boogie on the generated files
			-- delete all those that don't typecheck
			-- and rename them with `failed_name' prefix
		   	-- all those none of whose generated invariants can be
		   	-- checked by Boogie
		   	-- also return the number of correctly checked invariants across all
		   	-- files
		local
			a_file: PLAIN_TEXT_FILE
			new_invariants: INTEGER
		do
			Result := 0
			from mutated_files_names.start
			until mutated_files_names.after
			loop
				if not type_checks (mutated_files_names.item) then
					-- typechecking error: delete file
					create a_file.make (mutated_files_names.item)
					a_file.delete
					mutated_files_names.remove
				else
					new_invariants := verified_invariants (mutated_files_names.item)
					Result := Result + new_invariants
					if new_invariants = 0 then
						-- rename by adding `failed_name' as prefix
						create a_file.make (mutated_files_names.item)
						if a_file.is_writable then
							a_file.change_name (failed_name + mutated_files_names.item)
						end
					end
					mutated_files_names.forth
				end
				-- output some advancement indicator
				print ("%N" + "Boogie checked a file.")
			end
		end


	type_checks (a_file_name: STRING): BOOLEAN is
			-- launch typecheck (i.e., option /noVerify) of Boogie
			-- on a_file_name and return True iff typecheck yields no errors
		local
			arguments: LINKED_LIST[STRING]
			last_output: STRING
		do
			create arguments.make
			arguments.extend ("/noVerify")
			arguments.extend (a_file_name)
			last_output := launch_command (boogie_command_name, arguments)
			Result := last_output.has_substring ("Boogie program verifier finished with 0 verified, 0 errors")
		end


	verified_invariants (a_file_name: STRING): INTEGER is
			-- launch Boogie on `a_file_name' and returns the number of
			-- automatically generated invariants that are verified
		local
			arguments: LINKED_LIST[STRING]
			a_file: PLAIN_TEXT_FILE
			current_line_number: INTEGER
			line_numbers_of_invariants: LINKED_SET [INTEGER]
			last_output: STRING
			error_line: INTEGER
			string_error_line: STRING
			output_list: LINKED_LIST [STRING]
			repeat: BOOLEAN
		do
			create a_file.make (a_file_name)
			if not a_file.is_access_readable then
				Result := 0
			else
				a_file.open_read
				-- extract line numbers of automatically generated invariants
				create line_numbers_of_invariants.make
				from a_file.read_line;  current_line_number := 1
				until a_file.end_of_file
				loop
					if a_file.last_string.ends_with (weakener.comment_line) then
						line_numbers_of_invariants.extend (current_line_number)
					end
					a_file.read_line
					current_line_number := current_line_number + 1
				end
				a_file.close

				Result := line_numbers_of_invariants.count

				if Result /= 0 then
					-- actually launch Boogie
					arguments := split_string_into_words (boogie_parameter, ' ')
					arguments.extend (a_file_name)

					-- repeat until necessary
					from repeat := True
					until not repeat
					loop
						-- launch Boogie
						last_output := launch_command (boogie_command_name, arguments)
						-- analyze output to see if some invariant is not verified
						output_list := split_string_into_words (last_output, '%N')
						from output_list.start
						until output_list.after
						loop
							if output_list.item.has_substring (a_file_name + "(") and
								(output_list.item.has_substring ("Error BP5004") or
							 	 output_list.item.has_substring ("Error BP5005")) then
								string_error_line := output_list.item.substring (a_file_name.count + 2, output_list.item.index_of (',', a_file_name.count + 1) - 1)
								-- substring after a_file_name( and before comma should be a line number
								check string_error_line.is_integer end
								error_line := string_error_line.to_integer
								if line_numbers_of_invariants.has (error_line) then
									line_numbers_of_invariants.prune (error_line)
									Result := Result - 1
									-- rebuild the file with the unverified invariant commented out
									rebuild_file_with_commented_line (a_file_name, error_line)
									-- will re-launch Boogie on the same file
									output_list.finish
									repeat := True
								end
							else
								repeat := False
							end
							output_list.forth
						end
					end
				end
			end
		end


	split_string_into_words (a_string: STRING; a_whitespace: CHARACTER): LINKED_LIST [STRING] is
			-- split `a_string' into words (separated by a_whitespace) and
			-- put every word into the returned list
		local
			index: INTEGER
			modifiable_string: STRING
		do
			create modifiable_string.make_from_string (a_string)
			modifiable_string.left_adjust
			create Result.make
			index := 1

			from index := modifiable_string.index_of (a_whitespace, 1)
			until index < 1
			loop
				Result.extend (modifiable_string.substring (1, index - 1))
				modifiable_string.remove_head (index)
				modifiable_string.left_adjust
				index := modifiable_string.index_of (a_whitespace, 1)
			end
			if not modifiable_string.is_empty then
				Result.extend (modifiable_string)
			end
		end







feature {NONE} -- Program launching



	launch_command (command: STRING; arguments: LINKED_LIST[STRING]): STRING is
			-- launch `command' and
			-- return the output of the execution
		local
			p: PROCESS
			pf: PROCESS_FACTORY
		do
            create Result.make (10)
            create pf
            p := pf.process_launcher (command, arguments, Void)
            p.redirect_output_to_agent (agent (res: STRING; s: STRING)
                do
                    if s /= Void then
                        res.append_string (s)
                    end
                end (Result, ?)
                )
            p.launch
            p.wait_for_exit
        end



       rebuild_file_with_commented_line (a_file_name: STRING; line: INTEGER) is
       		-- rebuild file `a_file_name' with line number `line' commented out with "//"
       	local
       		a_file: PLAIN_TEXT_FILE
       		file_on_a_string: STRING
       		current_line: INTEGER
       	do
       		create a_file.make_open_read (a_file_name)
       		create file_on_a_string.make_empty

       		from a_file.read_line;  current_line := 1
       		until a_file.end_of_file
       		loop
       			if current_line = line then
       				file_on_a_string.append ("// ")
       			end
       			file_on_a_string.append (a_file.last_string)
       			file_on_a_string.extend ('%N')
       			current_line := current_line + 1
       			a_file.read_line
       		end
       		a_file.close

       		create a_file.make_open_write (a_file_name)
       		a_file.put_string (file_on_a_string)
       		a_file.close
       	end


end
