-- Copyright (C) 2009  Carlo A. Furia
--                     caf@inf.ethz.ch  http://se.inf.ethz.ch/people/furia/
--                     
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.

-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.

-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.


class
	MAPPINGS [D, C]


create
	make, make_with_restricted_domain



feature {NONE} -- Implementation

counter: ARRAYED_LIST [INTEGER]
	-- counter[i] = j means that domain element #i maps to
	-- codomain element #j (j = 0 means mapping is not defined)



feature {NONE} -- Initialization

	make (a_domain: LINKED_SET[D]; a_codomain: LINKED_SET[C]) is
		-- set domain and codomain sets
		require
			a_domain /= Void
			a_codomain /= Void
		do
			domain_set := a_domain
			domain_size := domain_set.count
			codomain_set := a_codomain
			codomain_size := codomain_set.count
			dom_elements_per_time := domain_size
		ensure
			domain_set /= Void
			codomain_set /= Void
		end


	make_with_restricted_domain (a_domain: LINKED_SET[D]; a_codomain: LINKED_SET[C]; number_of_domain_elements: INTEGER) is
		-- set domain and codomain sets and restrict the mappings to number_of_domain_elements elements in the domain
		require
			a_domain /= Void
			a_codomain /= Void
			number_of_domain_elements >= 0
		do
			make (a_domain, a_codomain)
			if number_of_domain_elements <= a_domain.count then
				dom_elements_per_time := number_of_domain_elements
			else
				dom_elements_per_time := a_domain.count
			end

		ensure
			domain_set /= Void
			codomain_set /= Void
			dom_elements_per_time <= number_of_domain_elements
		end




feature -- Mappings

	domain_set: LINKED_SET [D]
	domain_size: INTEGER

	codomain_set: LINKED_SET [C]
	codomain_size: INTEGER

	dom_elements_per_time: INTEGER


	start is
			-- initialize the enumeration of mappings
		do
			-- initialize mapping to all undefined
			create counter.make (domain_size)
			from domain_set.start
			until domain_set.after
			loop
				counter.extend (0)
				domain_set.forth
			end
			if domain_size = 0 then
				after := True
			else
				after := False
			end
		end



	after: BOOLEAN
		-- all mappings have been enumerated



	forth is
			-- move to next mapping
		local
			digit: INTEGER
			done: BOOLEAN
		do
			from digit := 1
			until (digit > domain_size) or (done and is_right_cardinality)
			loop
				counter [digit] := counter [digit] + 1
				if counter [digit] <= codomain_size then
					done := True
				else
					done := False
					counter [digit] := 0
					digit := digit + 1
				end
			end
			if digit > domain_size then
				after := True
			end
		end


	is_defined (item: D): BOOLEAN is
			-- is the mapping for element `item' defined in the current mapping?
		require
			domain_set.has (item)
		do
			Result := counter [domain_set.index_of (item, 1)] > 0
		end


	is_right_cardinality: BOOLEAN is
			-- is the current mapping of the right cardinality w.r.t. dom_elements_per_time?
		local
			zeros: INTEGER
		do
			zeros := 0
			from counter.start
			until counter.after
			loop
				if counter.item = 0 then
					zeros := zeros + 1
				end
				counter.forth
			end
			Result := (counter.count - zeros) <= dom_elements_per_time
		end




	maps_to (item: D): C is
			-- returns the element `item' maps to under the current mapping
			-- or Void if such mapping is undefined
		require
			is_defined (item)
		do
			Result := codomain_set [counter [domain_set.index_of (item, 1)]]
		end


end
